"""
hatch-pip-compile
"""

from hatch_pip_compile.__about__ import __application__, __author__, __email__, __version__

__all__ = [
    "__application__",
    "__version__",
    "__author__",
    "__email__",
]
